import math
from abc import abstractmethod

import pycode_similar
from diff_match_patch import diff_match_patch


class Comparator:

    @abstractmethod
    def compare(self, a, b):
        pass


class PycodeComparison(Comparator):

    def compare(self, a, b):
        res = pycode_similar.detect([a, b], diff_method=pycode_similar.UnifiedDiff)
        val = res[0][1][0].plagiarism_percent
        return val


class DiffComparator(Comparator):

    def __init__(self, base=None) -> None:
        super().__init__()

        self.H = None

        if base is not None:
            self.H = set()
            for e in base:
                self.H.add(e)

    def filter(self, s):
        return [e for e in s if e not in self.H]

    def compare(self, a, b):

        if self.H is not None:
            a, b = self.filter(a), self.filter(b)

        dmp = diff_match_patch()
        dmp.Diff_Timeout = math.inf
        ret = dmp.diff_linesToChars("\n".join(a), "\n".join(b))

        diff = dmp.diff_main(*ret)
        dmp.diff_charsToLines(diff, ret[2])
        #dmp.diff_cleanupSemantic(diff)

        perc = 1 - len([e for e in diff if e[0] != 0]) / (len(a) + len(b))

        #html = dmp.diff_prettyHtml(diff)
        #with open("test.html", "w") as f:
        #    f.write(html)

        return perc



import os
from collections import namedtuple

import numpy as np
import plotly.graph_objects as go

from pyplagiarism.vendor.diff2html.diff2html import CodeDiff


def parse(path_to_file, sourround_with_method=True):
    with open(path_to_file, "r") as f:
        code = f.readlines()

    if sourround_with_method:
        code = ["   " + line for line in code]
        code.insert(0, "def run():")

    return "\n".join(code)


def create_folder(p):
    if not os.path.exists(p):
        os.makedirs(p)


def get_files(path_to_folder):
    ret = []
    for root, dirs, files in os.walk(os.path.abspath(path_to_folder)):
        for f in files:
            ret.append(os.path.join(root, f))
    return sorted(ret)


def get_unique_key(data, f):
    name = os.path.basename(f)

    # remove the extension of necessary
    if '.' in name:
        name = '.'.join(name.split('.')[:-1])

    while name in data:
        name = "_" + name

    return name


def files_to_dict(*path_to_files):
    data = {}
    for f in path_to_files:
        key = get_unique_key(data, f)
        data[key] = parse(f)
    return data


def find_groups(labels, M, perc=0.95):
    b = np.full(len(labels), False)

    groups = []

    # a simple expand function which allows recursion
    def expand(i, b, g):
        if not b[i]:
            b[i] = True

            I = np.where(~b)[0]
            others = I[np.where(M[k, I] >= perc)[0]].tolist()

            for o in others:
                if not b[o]:
                    expand(o, b, g)
                    g.append(o)

    # until all students are considered
    while b.sum() < len(labels):

        I = np.where(~b)[0]
        k = int(I.min())

        val = [k]
        expand(k, b, val)

        if len(val) > 1:
            groups.append(val)

    groups = [sorted(group) for group in groups]

    return groups


def sort_by_plag(M, groups=None):
    S = []

    if groups is not None:
        for group in groups:
            S.extend(group)

    I = M.max(axis=1).argsort()[::-1]
    for i in I:
        if i not in S:
            S.append(int(i))

    return S


def plot(M, labels, **kwargs):
    val = np.copy(M)

    val = val ** 6
    val = 0.5 + (val / 2)

    np.fill_diagonal(val, np.inf)
    val[np.isinf(val)] = 0

    fig = go.FigureWidget(
        data=go.Heatmap(
            z=val,
            text=kwargs.get("P"),
            hovertext=M,
            x=labels,
            y=labels,
            colorscale=[
                [0, "grey"],
                [0.5, "white"],
                [1, "red"]],
            hoverinfo="x+y+z+text",
            zmin=0,
            zmax=1
        ))

    return fig


def visualize(out, M, labels, **kwargs):
    fig = plot(M, labels, **kwargs)

    html = fig.to_html()

    script = """
        <script>
            var myPlot =  document.getElementsByClassName("plotly-graph-div")[0];
            myPlot.on('plotly_click', function(data){
                var obj = data.points[0].text;
                if (obj.length > 0) {
                    window.open('diff/' + obj);
                }   
            });
        </script>
        """
    html += script

    with open(out, "w") as f:
        f.write(html)


def diff(out, data):
    labels = list(data.keys())
    P = [["" for _ in range(len(labels))] for _ in range(len(labels))]

    for i in range(len(labels)):
        for j in range(i + 1, len(labels)):
            file_a, file_b = data[labels[i]], data[labels[j]]
            Option = namedtuple("Options", "file1 file2 print_width show syntax_css verbose")
            options = Option(file1=file_a, file2=file_b, print_width=True, show=False, syntax_css="vs", verbose=False)

            name_a = os.path.basename(file_a)
            name_b = os.path.basename(file_b)

            name = f"{name_a}_vs_{name_b}".replace(".py", "")
            fname = name + ".html"
            path = os.path.join(out, fname)

            codeDiff = CodeDiff(None, None, fromtxt=file_a, totxt=file_b, name=name)
            codeDiff.format(options)
            codeDiff.write(path)

            P[i][j], P[j][i] = fname, fname

    return np.array(P)


def check_similarities(data, cmp, verbose=False):
    label = np.array(list(data.keys()), dtype=np.str)

    M = np.full((len(label), len(label)), np.inf)

    # now compare the input for each pair of label
    for i in range(len(label)):
        for j in range(i + 1, len(label)):
            path_a, path_b = label[i], label[j]

            a, b = data[path_a], data[path_b]

            try:
                val = cmp.compare(a, b)
            except:
                val = 0.0
                if verbose:
                    print(f"Error while comparing {path_a} with {path_b}")

            M[i, j], M[j, i] = val, val

    return label, M


def check_groups(labels, M, verbose=True):

    groups = find_groups(labels, M)

    if len(groups) == 0:
        print("No plagiarism groups were found!")
    else:
        for i in range(len(groups)):
            g = groups[i]
            val = np.round(M[g][:, g].min(), 2)
            s = ", ".join(labels[g])

            if verbose:
                print(f"Group {i + 1} ({val}): {s}")

    return groups



# MIT License
#
# Copyright (c) 2016 Alex Goodman
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
# of the Software, and to permit persons to whom the Software is furnished to do
# so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import io
import os
import sys
import difflib
import argparse
import pygments
import webbrowser

from pygments.formatters.html import HtmlFormatter
from pygments.lexers import guess_lexer_for_filename
from pygments.lexer import RegexLexer
from pygments.token import *

# Monokai is not quite right yet
from pyparsing import unicode

PYGMENTS_STYLES = ["vs", "xcode"]

HTML_TEMPLATE = """
<!DOCTYPE html>
<html class="no-js">
    <head>
        <!-- 
          html_title:    browser tab title
          reset_css:     relative path to reset css file
          pygments_css:  relative path to pygments css file
          diff_css:      relative path to diff layout css file
          page_title:    title shown at the top of the page. This should be the filename of the files being diff'd
          original_code: full html contents of original file
          modified_code: full html contents of modified file
          jquery_js:     path to jquery.min.js
          diff_js:       path to diff.js
        -->
        <meta charset="utf-8">
        <title>
            %(html_title)s
        </title>
        <meta name="description" content="">
        <meta name="viewport" content="width=device-width, initial-scale=1">
        <meta name="mobile-web-app-capable" content="yes">
        <link rel="stylesheet" href="%(reset_css)s" type="text/css">
        <link rel="stylesheet" href="%(diff_css)s" type="text/css">
        <link class="syntaxdef" rel="stylesheet" href="%(pygments_css)s" type="text/css">
    </head>
    <body>
        <div class="" id="topbar">
          <div id="filetitle"> 
            %(page_title)s
          </div>
          <div class="switches">
            <div class="switch">
              <input id="showoriginal" class="toggle toggle-yes-no menuoption" type="checkbox" checked>
              <label for="showoriginal" data-on="&#10004; Original" data-off="Original"></label>
            </div>
            <div class="switch">
              <input id="showmodified" class="toggle toggle-yes-no menuoption" type="checkbox" checked>
              <label for="showmodified" data-on="&#10004; Modified" data-off="Modified"></label>
            </div>
            <div class="switch">
              <input id="highlight" class="toggle toggle-yes-no menuoption" type="checkbox" checked>
              <label for="highlight" data-on="&#10004; Highlight" data-off="Highlight"></label>
            </div>
            <div class="switch">
              <input id="codeprintmargin" class="toggle toggle-yes-no menuoption" type="checkbox" checked>
              <label for="codeprintmargin" data-on="&#10004; Margin" data-off="Margin"></label>
            </div>
            <div class="switch">
              <input id="dosyntaxhighlight" class="toggle toggle-yes-no menuoption" type="checkbox" checked>
              <label for="dosyntaxhighlight" data-on="&#10004; Syntax" data-off="Syntax"></label>
            </div>
          </div>
        </div>
        <div id="maincontainer" class="%(page_width)s">
            <div id="leftcode" class="left-inner-shadow codebox divider-outside-bottom">
                <div class="codefiletab">
                    &#10092; Original
                </div>
                <div class="printmargin">
                    01234567890123456789012345678901234567890123456789012345678901234567890123456789
                </div>
                %(original_code)s
            </div>
            <div id="rightcode" class="left-inner-shadow codebox divider-outside-bottom">
                <div class="codefiletab">
                    &#10093; Modified
                </div>
                <div class="printmargin">
                    01234567890123456789012345678901234567890123456789012345678901234567890123456789
                </div>
                %(modified_code)s
            </div>
        </div>
<script src="%(jquery_js)s" type="text/javascript"></script>
<script src="%(diff_js)s" type="text/javascript"></script>
    </body>
</html>
"""


class DefaultLexer(RegexLexer):
    """
    Simply lex each line as a token.
    """

    name = 'Default'
    aliases = ['default']
    filenames = ['*']

    tokens = {
        'root': [
            (r'.*\n', Text),
        ]
    }


class DiffHtmlFormatter(HtmlFormatter):
    """
    Formats a single source file with pygments and adds diff highlights based on the 
    diff details given.
    """
    isLeft = False
    diffs = None

    def __init__(self, isLeft, diffs, *args, **kwargs):
        self.isLeft = isLeft
        self.diffs = diffs
        super(DiffHtmlFormatter, self).__init__(*args, **kwargs)

    def wrap(self, source, outfile):
        return self._wrap_code(source)

    def getDiffLineNos(self):
        retlinenos = []
        for idx, ((left_no, left_line), (right_no, right_line), change) in enumerate(self.diffs):
            no = None
            if self.isLeft:
                if change:
                    if isinstance(left_no, int) and isinstance(right_no, int):
                        no = '<span class="lineno_q lineno_leftchange">' + \
                            str(left_no) + "</span>"
                    elif isinstance(left_no, int) and not isinstance(right_no, int):
                        no = '<span class="lineno_q lineno_leftdel">' + \
                            str(left_no) + "</span>"
                    elif not isinstance(left_no, int) and isinstance(right_no, int):
                        no = '<span class="lineno_q lineno_leftadd">  </span>'
                else:
                    no = '<span class="lineno_q">' + str(left_no) + "</span>"
            else:
                if change:
                    if isinstance(left_no, int) and isinstance(right_no, int):
                        no = '<span class="lineno_q lineno_rightchange">' + \
                            str(right_no) + "</span>"
                    elif isinstance(left_no, int) and not isinstance(right_no, int):
                        no = '<span class="lineno_q lineno_rightdel">  </span>'
                    elif not isinstance(left_no, int) and isinstance(right_no, int):
                        no = '<span class="lineno_q lineno_rightadd">' + \
                            str(right_no) + "</span>"
                else:
                    no = '<span class="lineno_q">' + str(right_no) + "</span>"

            retlinenos.append(no)

        return retlinenos

    def _wrap_code(self, source):
        source = list(source)
        yield 0, '<pre>'

        for idx, ((left_no, left_line), (right_no, right_line), change) in enumerate(self.diffs):
            # print idx, ((left_no, left_line),(right_no, right_line),change)
            try:
                if self.isLeft:
                    if change:
                        if isinstance(left_no, int) and isinstance(right_no, int) and left_no <= len(source):
                            i, t = source[left_no - 1]
                            t = '<span class="left_diff_change">' + t + "</span>"
                        elif isinstance(left_no, int) and not isinstance(right_no, int) and left_no <= len(source):
                            i, t = source[left_no - 1]
                            t = '<span class="left_diff_del">' + t + "</span>"
                        elif not isinstance(left_no, int) and isinstance(right_no, int):
                            i, t = 1, left_line
                            t = '<span class="left_diff_add">' + t + "</span>"
                        else:
                            raise
                    else:
                        if left_no <= len(source):
                            i, t = source[left_no - 1]
                        else:
                            i = 1
                            t = left_line
                else:
                    if change:
                        if isinstance(left_no, int) and isinstance(right_no, int) and right_no <= len(source):
                            i, t = source[right_no - 1]
                            t = '<span class="right_diff_change">' + t + "</span>"
                        elif isinstance(left_no, int) and not isinstance(right_no, int):
                            i, t = 1, right_line
                            t = '<span class="right_diff_del">' + t + "</span>"
                        elif not isinstance(left_no, int) and isinstance(right_no, int) and right_no <= len(source):
                            i, t = source[right_no - 1]
                            t = '<span class="right_diff_add">' + t + "</span>"
                        else:
                            raise
                    else:
                        if right_no <= len(source):
                            i, t = source[right_no - 1]
                        else:
                            i = 1
                            t = right_line
                yield i, t
            except:
                # print "WARNING! failed to enumerate diffs fully!"
                pass  # this is expected sometimes
        yield 0, '\n</pre>'

    def _wrap_tablelinenos(self, inner):
        dummyoutfile = io.StringIO()
        lncount = 0
        for t, line in inner:
            if t:
                lncount += 1

            # compatibility Python v2/v3
            if sys.version_info > (3,0):
                dummyoutfile.write(line)
            else:
                dummyoutfile.write(unicode(line))

        fl = self.linenostart
        mw = len(str(lncount + fl - 1))
        sp = self.linenospecial
        st = self.linenostep
        la = self.lineanchors
        aln = self.anchorlinenos
        nocls = self.noclasses

        lines = []
        for i in self.getDiffLineNos():
            lines.append('%s' % (i,))

        ls = ''.join(lines)

        # in case you wonder about the seemingly redundant <div> here: since the
        # content in the other cell also is wrapped in a div, some browsers in
        # some configurations seem to mess up the formatting...
        if nocls:
            yield 0, ('<table class="%stable">' % self.cssclass +
                      '<tr><td><div class="linenodiv" '
                      'style="background-color: #f0f0f0; padding-right: 10px">'
                      '<pre style="line-height: 125%">' +
                      ls + '</pre></div></td><td class="code">')
        else:
            yield 0, ('<table class="%stable">' % self.cssclass +
                      '<tr><td class="linenos"><div class="linenodiv"><pre>' +
                      ls + '</pre></div></td><td class="code">')
        yield 0, dummyoutfile.getvalue()
        yield 0, '</td></tr></table>'


class CodeDiff(object):
    """
    Manages a pair of source files and generates a single html diff page comparing
    the contents.
    """
    root = os.path.dirname(os.path.realpath(__file__))


    pygmentsCssFile = f"{root}/deps/codeformats/%s.css"
    diffCssFile = f"{root}/deps/diff.css"
    diffJsFile = f"{root}/deps/diff.js"
    resetCssFile = f"{root}/deps/reset.css"
    jqueryJsFile = f"{root}/deps/jquery.min.js"

    def __init__(self, fromfile, tofile, fromtxt=None, totxt=None, name=None):
        self.filename = name
        self.fromfile = fromfile
        if fromtxt == None:
            try:
                with io.open(fromfile) as f:
                    self.fromlines = f.readlines()
            except Exception as e:
                print("Problem reading file %s" % fromfile)
                print(e)
                sys.exit(1)
        else:
            self.fromlines = [n + "\n" for n in fromtxt.split("\n")]
        self.leftcode = "".join(self.fromlines)

        self.tofile = tofile
        if totxt == None:
            try:
                with io.open(tofile) as f:
                    self.tolines = f.readlines()
            except Exception as e:
                print("Problem reading file %s" % tofile)
                print(e)
                sys.exit(1)
        else:
            self.tolines = [n + "\n" for n in totxt.split("\n")]
        self.rightcode = "".join(self.tolines)

    def getDiffDetails(self, fromdesc='', todesc='', context=False, numlines=5, tabSize=8):
        # change tabs to spaces before it gets more difficult after we insert
        # markkup
        def expand_tabs(line):
            # hide real spaces
            line = line.replace(' ', '\0')
            # expand tabs into spaces
            line = line.expandtabs(tabSize)
            # replace spaces from expanded tabs back into tab characters
            # (we'll replace them with markup after we do differencing)
            line = line.replace(' ', '\t')
            return line.replace('\0', ' ').rstrip('\n')

        self.fromlines = [expand_tabs(line) for line in self.fromlines]
        self.tolines = [expand_tabs(line) for line in self.tolines]

        # create diffs iterator which generates side by side from/to data
        if context:
            context_lines = numlines
        else:
            context_lines = None

        diffs = difflib._mdiff(self.fromlines, self.tolines, context_lines,
                               linejunk=None, charjunk=difflib.IS_CHARACTER_JUNK)
        return list(diffs)

    def format(self, options):
        self.diffs = self.getDiffDetails(self.fromfile, self.tofile)

        if options.verbose:
            for diff in self.diffs:
                print("%-6s %-80s %-80s" % (diff[2], diff[0], diff[1]))

        fields = ((self.leftcode, True, self.fromfile),
                  (self.rightcode, False, self.tofile))

        codeContents = []
        for (code, isLeft, filename) in fields:

            inst = DiffHtmlFormatter(isLeft,
                                     self.diffs,
                                     nobackground=False,
                                     linenos=True,
                                     style=options.syntax_css)

            try:
                self.lexer = guess_lexer_for_filename(self.filename, code)

            except pygments.util.ClassNotFound:
                if options.verbose:
                    print("No Lexer Found! Using default...")

                self.lexer = DefaultLexer()

            formatted = pygments.highlight(code, self.lexer, inst)

            codeContents.append(formatted)

        answers = {
            "html_title":     self.filename,
            "reset_css":      self.resetCssFile,
            "pygments_css":   self.pygmentsCssFile % options.syntax_css,
            "diff_css":       self.diffCssFile,
            "page_title":     self.filename,
            "original_code":  codeContents[0],
            "modified_code":  codeContents[1],
            "jquery_js":      self.jqueryJsFile,
            "diff_js":        self.diffJsFile,
            "page_width":     "page-80-width" if options.print_width else "page-full-width"
        }

        self.htmlContents = HTML_TEMPLATE % answers

    def write(self, path):
        fh = io.open(path, 'w')
        fh.write(self.htmlContents)
        fh.close()


def main(file1, file2, outputpath, options):
    codeDiff = CodeDiff(file1, file2, name=file2)
    codeDiff.format(options)
    codeDiff.write(outputpath)

def show(outputpath):
    path = os.path.abspath(outputpath)
    webbrowser.open('file://' + path)



if __name__ == "__main__":
    description = """Given two source files this application\
creates an html page which highlights the differences between the two. """

    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('-s', '--show', action='store_true',
                        help='show html in a browser.')
    parser.add_argument('-p', '--print-width', action='store_true',
        help='Restrict code to 80 columns wide. (printer friendly in landscape)')
    parser.add_argument('-c', '--syntax-css', action='store', default="vs",
        help='Pygments CSS for code syntax highlighting. Can be one of: %s' % str(PYGMENTS_STYLES))
    parser.add_argument('-v', '--verbose', action='store_true', help='show verbose output.')
    parser.add_argument('file1', help='source file to compare ("before" file).')
    parser.add_argument('file2', help='source file to compare ("after" file).')

    args = parser.parse_args()

    if args.syntax_css not in PYGMENTS_STYLES:
        raise ValueError("Syntax CSS (-c) must be one of %r." % PYGMENTS_STYLES)

    outputpath = "index.html"
    main(args.file1, args.file2, outputpath, args)
    if args.show:
        show(outputpath)



import os
import webbrowser

from pyplagiarism.comparator import PycodeComparison
from pyplagiarism.util import get_files, files_to_dict, sort_by_plag, diff, check_similarities, check_groups, visualize, \
    create_folder


def plagiarism(data,
               output_folder=None,
               comparator=PycodeComparison(),
               find_groups=True,
               visualize_as_html=True,
               sort_by_plagiarism=True,
               diff_of_files=True,
               verbose=True
               ):
    """

    Function that does the job of plagiarism check.

    Parameters
    ----------

    data
        A dictionary where each entry is a source code file.

    output_folder
        The output folder where the results should be stored

    find_groups
        Whether groups of similarities should be detected.

    visualize_as_html
        Whether the output should be written to file file that visualizes the plagiarism in a
        matrix.

    sort_by_plagiarism
        Whether the files should be sorted by corresponding amount of plagiarism when visualized in html.
        Otherwise, the files are sorted by alphabet.

    diff_of_files
        Whether for each pair of files a diff should be created.

    comparator
        The comparator which returns compares two lists of strings (the lines of each file) and
        returns the corresponding plagiarism value. New comparators can be written and provided.

    verbose
        Whether output should be printed or not.

    """

    if output_folder is None:
        if visualize_as_html or diff_of_files:
            raise Exception("Please define output_folder if the results should be visualized or diff files written.")
    else:
        create_folder(output_folder)

    # actually run the comparisons
    labels, M = check_similarities(data, comparator, verbose=verbose)

    groups = None
    path_to_diff = None

    if find_groups:
        groups = check_groups(labels, M)

    if diff_of_files:
        diff_path = os.path.join(output_folder, "diff")
        create_folder(diff_path)
        path_to_diff = diff(diff_path, data)

    if sort_by_plagiarism:
        I = sort_by_plag(M, groups=groups)
        M = M[I][:, I]
        labels = labels[I]
        if path_to_diff is not None:
            path_to_diff = path_to_diff[I][:, I]

    if visualize_as_html:
        out_index = os.path.join(output_folder, "index.html")
        visualize(out_index, M, labels, P=path_to_diff)
        webbrowser.open('file://' + out_index)


def plagiarism_from_files(*files):
    # if just one entry we assume it is the folder
    if len(files) == 1:
        files = get_files(files[0])

    # parse all the data into an array
    return files_to_dict(*files)



from setuptools import setup, find_packages
from pyplagiarism.version import __version__

def readme():
    with open('README.rst') as f:
        return f.read()


__name__ = "pyplagiarism"
__author__ = "Julian Blank"
__url__ = "https://github.com/julesy89/pyplagiarism"

setup(
    name=__name__,
    version=__version__,
    author=__author__,
    author_email="blankjul@egr.msu.edu",
    description="Plagiarism Tool for Source Code",
    long_description=readme(),
    url=__url__,
    license='Apache License 2.0',
    keywords="optimization",
    install_requires=["numpy", "pycode_similar", "plotly", "mistune", "pygments", "diff-match-patch"],
    packages=find_packages(exclude=['tests', 'doc']),
    include_package_data=True,
    platforms='any',
)



from pyplagiarism.tool import plagiarism

a = """
def calculate():
    x = 6
    y = 5
    return x + y
"""

b = """
def calculate():
    return 5 + 6
"""

c = """
def calculate():
    x = 6
    y = 5
    return x + y
"""

data = {"a": a, "b": b, "c": c}

plagiarism(data, visualize_as_html=False, diff_of_files=False)
